// -*- mode:C++ -*-
//Header:
//Title: Sample of Soccer Client Program
//File: client.h
//Author: NODA, Itsuki
//Date: 1996/05/19
//EndHeader:

//======================================================================
//Parameters
//======================================================================
#define NAMEMAX	1024
#define BUFFERSIZE 4096
#define MINIBUFSIZE 128
#define DEFAULT_TEAMNAME "ETL"
#define DEFAULT_SERVER_HOST "localhost"
#define DEFAULT_SERVER_PORT 6000
#define DEFAULT_POSITION PS_dango
#define DEFAULT_POSITION_STR "dango"
#define VISUAL_OBJ_TABLE_SIZE 64
#define EPS 0.00001
#define RECV_INTERVAL_MILISEC	100

#define RAD2DEG (180.0 / M_PI) 
#define DEG2RAD (M_PI / 180.0) 

#define DefaultKickMaxPow 40.0
#define DefaultKickOffset 20.0


//======================================================================
//Types
//======================================================================
typedef double Real ;
typedef int Int ;
typedef char* Buffer ;
enum Boolean {False, True} ;
class PPos {
public:  
     Real x ; 
     Real y ;

     inline friend ostream& operator<<(ostream& ost, const PPos& pos) {
	  return ost << "#Pos[" << pos.x << "," << pos.y << "]" ;
     }
     inline PPos(Real vx = 0.0, Real vy = 0.0) {
	  x = vx ; y = vy ;
     } ;
     
     inline Real distance(PPos pos) {
       Real dx = x - pos.x ;
       Real dy = y - pos.y ;
       Real d = sqrt(dx * dx + dy * dy) ;
       return d ;
     } ;
     inline Real distance(PPos pos, PPos r) {
       Real dx = (x - pos.x)/r.x ;
       Real dy = (y - pos.y)/r.y ;
       return sqrt(dx * dx + dy * dy) ;
     } ;
} ;
       

enum ViewAngle {VA_Narrow, VA_Normal, VA_Wide} ;
String ViewAngleStr[] = {"narrow","normal","wide"} ;
enum ViewQuality {VQ_High, VQ_Low} ;
String ViewQualityStr[] = {"high","low"} ;

enum LandMarkID {
     LM_ball,
     LM_goal,
     LM_owngoal,
     LM_flag_fl,
     LM_flag_fr,
     LM_flag_cl,
     LM_flag_cr,
     LM_flag_bl,
     LM_flag_br,
     LM_flag_pfl,
     LM_flag_pfc,
     LM_flag_pfr,
     LM_flag_pbl,
     LM_flag_pbc,
     LM_flag_pbr,
     LM_max} ;

PPos LandMarkPos[] = {
     PPos(  0.0,  0.0),
     PPos( 52.5,  0.0),
     PPos(-52.5,  0.0),
     PPos( 52.5,-34.0),
     PPos( 52.5, 34.0),
     PPos(  0.0,-34.0),
     PPos(  0.0, 34.0),
     PPos(-52.5,-34.0),
     PPos(-52.5, 34.0),
     PPos( 36.0,-20.16),
     PPos( 36.0,  0.0),
     PPos( 36.0, 20.16),
     PPos(-36.0,-20.16),
     PPos(-36.0,  0.0),
     PPos(-36.0, 20.16),
     PPos(  0.0,  0.0)
     } ;

enum Side { Side_L, Side_R, Side_none } ;

const String Side_L_str = "l" ;
const String Side_R_str = "r" ;

inline Side str_to_Side(String str) { 
     return str_to_Side((const char*)str) ;
}
inline Side str_to_Side(const char* str) {
     if(Side_L_str == str) 
	  return Side_L ;
     else if(Side_R_str == str)
	  return Side_R ;
     else 
	  return Side_none ;
}

String LandMarkStrL[] = {
     "ball","goal r","goal l",
     "flag r t","flag r b",
     "flag c t","flag c b",
     "flag l t","flag l b",
     "flag p r t","flag p r c","flag p r b",
     "flag p l t","flag p l c","flag p l b"} ; 
String LandMarkStrR[] = {
     "ball","goal l","goal r",
     "flag l b","flag l t",
     "flag c b","flag c t",
     "flag r b","flag r t",
     "flag p l b","flag p l c","flag p l t",
     "flag p r b","flag p r c","flag p r t"} ; 

enum PlayStyle {
     PS_dango,
     PS_nao,
     PS_nao_midfielder,
     PS_nao_defender,
     PS_top,
     PS_forward,
     PS_midfielder,
     PS_left_wingback,
     PS_right_wingback,
     PS_defender,
     PS_goalkeeper,
     PS_allrounder,
     PS_okano,
     PS_takeda,
     PS_jo,
     PS_maezono,
     PS_michiki,
     PS_narahashi,
     PS_nakata,
     PS_hattori,
     PS_ito,
     PS_endo,
     PS_tanaka,
     PS_suzuki,
     PS_matsuda,
     PS_kawaguchi,
     PS_GK,
     PS_LCB,
     PS_RCB,
     PS_LWB,
     PS_RWB,
     PS_BR,
     PS_OMF,
     PS_LW,
     PS_RW,
     PS_LCFW,
     PS_RCFW,
     PS_dummy,
     PS_max } ;
String PlayStyleStr[] = {
     "dango",
     "nao",
     "nao_midfielder",
     "nao_defender",
     "top",
     "forward",
     "midfielder",
     "left_wingback",
     "right_wingback",
     "defender",
     "goalkeeper",
     "allrounder",
     "okano",
     "takeda",
     "jo",
     "maezono",
     "michiki",
     "narahashi",
     "nakata",
     "hattori",
     "ito",
     "endo",
     "tanaka",
     "suzuki",
     "matsuda",
     "kawaguchi",
     "GK",
     "LCB",
     "RCB",
     "LWB",
     "RWB",
     "BR",
     "OMF",
     "LW",
     "RW",
     "LCFW",
     "RCFW",
     "dummy"} ;
Real PlayStyleRange[][4]  ={
     {-10,-30,-20, 20}, //dango
     {-10,-30,-20, 20}, //nao
     {-20,-35,-30, 30}, //nao_midfielder
     {-35,-45,-10, 10}, //nao_defender
     { -8,-20,-10, 10}, //top
     { -8,-20,-20, 20}, //forward
     {-20,-35,-30, 30}, //midfielder
     {-25,-40,-30,-10}, //left_wingback
     {-25,-40, 10, 30}, //right_wingback
     {-35,-45,-10, 10}, //defender
     {-45,-50, -5,  5}, //goalkeeper
     {-25,-35, -5,  5}, //allrounder
     { -8,-10, -5,  5}, //okano
     { -8,-10,-10, 10}, //takeda
     { -8,-10,-10, 10}, //jo
     {-15,-25,-10, 10}, //maezono
     {-25,-40,-30,-10}, //michiki
     {-25,-40, 30, 10}, //narahashi
     {-25,-30, -5,  5}, //nakata
     {-25,-35, -5,  5}, //hattori
     {-25,-35,-15, -5}, //ito
     {-25,-35,  5, 15}, //endo
     {-35,-45, -5,  5}, //tanaka
     {-35,-45,-15, -5}, //suzuki
     {-35,-45,  5, 15}, //matsuda
     {-50,-51, -2,  2}, //kawaguchi
     {-50,-51, -2,  2},//GK
     {-40,-45,-10,  0},//LCB
     {-40,-45,  0, 10},//RCB
     {-35,-40,-30,-20},//LWB
     {-35,-40, 20, 30},//RWB
     {-30,-35,-10, 10},//BR
     {-20,-25,-10, 10},//OMF
     {-5,-10,-30,-20},//LW
     {-5,-10, 20, 30},//RW
     { -8,-10,-10,  0},//LCFW
     { -8,-10,  0, 10},//RCFW
     {  0,-52,-34, 34}} ; //dummy

#define PM_GK	PPos(-52.0,  0.0) 
#define PM_LCB	PPos(-35.0, -8.0) 
#define PM_RCB	PPos(-35.0,  8.0) 
#define PM_LWB	PPos(-20.0,-24.0) 
#define PM_RWB	PPos(-20.0, 24.0) 
#define PM_BR	PPos(-10.0,  0.0) 
#define PM_OMF	PPos( 10.0,  0.0) 
#define PM_LW	PPos( 20.0,-24.0) 
#define PM_RW	PPos( 20.0, 24.0) 
#define PM_LCFW	PPos( 40.0, -8.0) 
#define PM_RCFW	PPos( 40.0,  8.0) 
#define PM_GL	PPos( 53.0,  0.0) 
#define PM_GLL	PPos( 53.0, -5.0) 
#define PM_GLR	PPos( 53.0,  5.0) 


#define RANDOMBASE 1000

inline Int randomize() {
     Int l ;
     static struct timeval Tv ;
     static struct timezone Tz ;
     gettimeofday( &Tv , &Tz ) ;
     l = (unsigned int)(Tv.tv_usec) ;
     srand( l ) ;
     return l ;
}     

inline Int irand(Int max) {
  return ((rand() % RANDOMBASE) % max) ;
}

inline Real drand(Real max) {
     return (max * ((double)(rand() % RANDOMBASE)
		    / (double)RANDOMBASE)) ;
}
inline Real drand(Real min, Real max) {
     return min + drand((max - min)) ;
}

inline Real normalize_angle(Real ang) {
     while(ang >  180.0) ang -= 360.0 ;
     while(ang < -180.0) ang += 360.0 ;
     return ang ;
}

inline Real Sin(Real x) { return sin(DEG2RAD * x) ; }
inline Real Cos(Real x) { return cos(DEG2RAD * x) ; }
inline Real Atan2(Real y, Real x) { return RAD2DEG * atan2(y,x) ; }
     

//======================================================================
//Function Prototypes
//======================================================================
void usage() ; 

//======================================================================
//Class Visual Object
//======================================================================
class VisualObject {
public:
     String name ;
     Real dist ;
     Real dir ;
     Int vdir ;
     Int time ;
     PPos pos ;
     Boolean fixp ;
} ;

//======================================================================
//Class Player
//======================================================================
class Player {
public:
     Socket sock ;
     String team ;
     String serverhost ;
     Int  serverport ;
     Side side ;
     Int  unum ;
     String play_style_str ;
     PlayStyle play_style ;
     Real cdir ;
     ViewAngle cva ;
     ViewQuality cvq ; 

     Int bufsize ;
     Buffer send_buffer ;
     Buffer recv_buffer ;

     Int clock ;
     Int time ;
     Int vtime ;
     Int atime ;
     VisualObject vobj[VISUAL_OBJ_TABLE_SIZE] ;
     Int N_vobj ;

     VisualObject* landmark[LM_max] ;
     VisualObject* ball ;
     VisualObject* goal ;
     VisualObject* owngoal ;

     PPos pos ;
     Real edir ; /* estimate direction */

     inline Player(const String teamname = DEFAULT_TEAMNAME,
		   const String host = DEFAULT_SERVER_HOST,
		   const Int port = DEFAULT_SERVER_PORT) {
	  team = teamname ;
	  serverhost = host ;
	  serverport = port ;
	  bufsize = BUFFERSIZE ;
	  send_buffer = new char[bufsize] ;
	  recv_buffer = new char[bufsize] ;
	  side = Side_L ;
	  N_vobj = 0 ;
	  time = 0 ;
	  cdir = 0.0 ;
	  cva = VA_Normal ;
	  cvq = VQ_High ;
	  play_style = DEFAULT_POSITION ;
	  play_style_str = DEFAULT_POSITION_STR ;
	  clock = time = 0 ;
     } ;

     Boolean init(int argc, char** argv) ;
     void mainloop() ;
     void close() ;

     void mygetopt(int argc, char** argv) ;

     void check_message() ;
     void start_sensor() ;
     void analyze_message() ;

     void voidaction() ;

     inline VisualObject* findVisualObject(char* name) {
	  return findVisualObject(String(name)) ;
     } ;
     VisualObject* findVisualObject(String name) ;
     inline void renewVisualObject(char* name, Real dist, Real dir, Int vdir,
				   Int type) {
	  renewVisualObject(String(name),dist,dir,vdir,type) ;
     } ;
     void renewVisualObject(String name, Real dist, Real dir, Int vdir,
			    Int type) ;
     void showStatus() ;

     PlayStyle str_to_position(String PosName) ;
     void move(PlayStyle pos) ;
     void move(Real x, Real y) ;
     void turn(Real moment) ;
     void turn_to(Real adir) ;
     void dash(Real power) ;
     inline void Dash(Real power, Real maxpow, Real offset) {
       if(power > (maxpow - offset))
	 dash(maxpow) ;
      else
	 dash(power + offset) ;
  } ;
     void kick(Real power, Real dir) ;
     inline void Kick(Real power, Real dir, Real maxpow, Real offset) {
       if(power > (maxpow - offset))
	 kick(maxpow,dir) ;
       else
	 kick(power + offset,dir) ;
     } ;
     void say(Buffer msg) ;
     void change_view(ViewAngle va, ViewQuality vq) ;

     inline void _pause(Int i) {
	  Int c = clock ; 
	  while(clock - c < i) {} 
     }
     inline void spause() { _pause(1) ; } ;
     inline void npause() { _pause(2) ; } ;
     inline void lpause() { _pause(5) ; } ;
     inline void Lpause() { _pause(10) ; } ;

     void wait_new_vinfo() ;
     Boolean new_vinfo_p(VisualObject* obj) ;
     Boolean kickable() ;
	  
     void look_around() ;
     void look_around_half() ;
     void look_at(String name) ;
     void look_at(VisualObject* name) ;
     void turn_to(String name) ;
     void turn_to(VisualObject* name) ;
     void search_obj(VisualObject* obj) ;

     void chase(String name, Real maxpow = 30.0, Real offset = 10.0 ) ;
     void chase(VisualObject* name, Real maxpow = 30.0, Real offset = 10.0 ) ;
     void chase_2(VisualObject* name) ;
     void chase_3(VisualObject* name) ;
     void chase_4(VisualObject* name) ;
     void go_toward(PPos tpos, Real maxpow = 20.0, Real offset = 10.0 ) ;

     void estimate_position(Boolean check_old_p = True) ;
     PPos estimate_object_pos(VisualObject* obj) ;

     void shoot_to_goal() ;
     void kick_to(VisualObject* obj, 
		  Real maxpow = DefaultKickMaxPow,
		  Real offset = DefaultKickOffset) ;
     void kick_immediately(VisualObject* obj,
			   Real maxpow = DefaultKickMaxPow,
			   Real offset = DefaultKickOffset) ;
     void kick_immediately(PPos tpos,
			   Real maxpow = DefaultKickMaxPow,
			   Real offset = DefaultKickOffset) ;
     void play_as_dango(); 
     void play_as_nao() ;
     void play_as_nao_midfielder() ;
     void play_as_nao_defender() ;
     void play_as_defender() ;

     void keep_defensive_pos() ;
     Boolean teammatep(VisualObject *obj) ;
     VisualObject* newest_fore_teammate() ;
     void clear_or_pass() ;
     void active_pass() ;

     Real find_clear_way(Real ddir, Real dist) ;
     Boolean clearwayp(Real to_dir,Real to_dist) ;
     Boolean playerp(VisualObject *obj) ;

     Boolean nearest_to_ballp() ;
     void play_as_midfielder() ;
     void shoot_or_pass() ;
     void offensive_pass() ;
     void wandering(PPos center, Real radius) ;

     void play_as_forward() ;
     void play_as_forward2() ;

     void wandering2(PPos center, Real rx, Real ry) ;
     void play_as_midfielder2(PPos center, Real rx, Real ry) ;

     void okano_dash(VisualObject* name) ;
     void play_as_okano() ;
     void play_as_takeda() ;
     void play_as_jo() ;

     void dribble_to(VisualObject* name) ;
     void power_dribble_to(VisualObject* name) ;
     void play_as_maezono() ;

     void centering(Real pow) ;
     void play_as_michiki() ;
     void play_as_narahashi() ;


     void through_pass() ;
     void play_as_nakata() ;

     void go_toward_backward(PPos tpos);

     inline void play_as_hattori() { 
	  play_as_midfielder2(PPos(-20, 15),30,20) ; 
     } ;
     inline void play_as_ito() { 
	  play_as_midfielder2(PPos(-20, 0),30,20) ; 
     } ;
     inline void play_as_endo() { 
	  play_as_midfielder2(PPos(-20, -15),30,20) ; 
     } ;

     inline void play_as_tanaka() { play_as_defender() ; };
     inline void play_as_suzuki() { play_as_defender() ; };
     inline void play_as_matsuda() { play_as_defender() ; };

     void play_as_kawaguchi() ;

     void play_as_relayer(PPos center, PPos r, PPos* dest, Int Ndist) ;

} ;
     
